unit URecordEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, axctrls, activex, ComObj, ExtCtrls,
  ComCtrls, ToolWin, ActnList, contnrs, ImgList, UEditDialog, UBiffEditUtil, Clipbrd,
  Menus, UBiffData, UWaitCursor, USearchDialog;

type
  TRecordEdit = class(TForm)
    ToolBar: TToolBar;
    ListRecords: TListView;
    Splitter1: TSplitter;
    btnExport: TToolButton;
    ExportDialog: TSaveDialog;
    ToolButton1: TToolButton;
    btnSearch: TToolButton;
    ToolButton2: TToolButton;
    btnEdit: TToolButton;
    btnUp: TToolButton;
    btnDown: TToolButton;
    btnDelete: TToolButton;
    ActionList: TActionList;
    ActionEdit: TAction;
    ActionMoveUp: TAction;
    ActionMoveDown: TAction;
    ActionDelete: TAction;
    ActionExport: TAction;
    ImageList1: TImageList;
    ImageList3: TImageList;
    ImageList2: TImageList;
    ActionSearch: TAction;
    ActionInsert: TAction;
    ActionCopy: TAction;
    ActionPaste: TAction;
    ToolButton5: TToolButton;
    ToolButton4: TToolButton;
    PopupMenu1: TPopupMenu;
    Copy1: TMenuItem;
    Paste1: TMenuItem;
    N1: TMenuItem;
    Edit1: TMenuItem;
    Delete1: TMenuItem;
    N2: TMenuItem;
    MoveUp1: TMenuItem;
    MoveDown1: TMenuItem;
    ActionCut: TAction;
    Cut1: TMenuItem;
    StatusBar: TStatusBar;
    btnOpen: TToolButton;
    btnSaveAs: TToolButton;
    ImagesListBox: TImageList;
    edHex: TRichEdit;
    procedure ListRecordsSelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure ListRecordsCustomDrawItem(Sender: TCustomListView;
      Item: TListItem; State: TCustomDrawState; var DefaultDraw: Boolean);
    procedure ListRecordsDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure ListRecordsDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure ActionExportExecute(Sender: TObject);
    procedure ActionDeleteExecute(Sender: TObject);
    procedure ActionEditExecute(Sender: TObject);
    procedure ActionEditUpdate(Sender: TObject);
    procedure ActionMoveUpExecute(Sender: TObject);
    procedure ActionMoveUpUpdate(Sender: TObject);
    procedure ActionDeleteUpdate(Sender: TObject);
    procedure ActionMoveDownExecute(Sender: TObject);
    procedure ActionMoveDownUpdate(Sender: TObject);
    procedure ActionSearchExecute(Sender: TObject);
    procedure ActionExportUpdate(Sender: TObject);
    procedure ListRecordsDblClick(Sender: TObject);
    procedure ActionSearchUpdate(Sender: TObject);
    procedure ActionCopyExecute(Sender: TObject);
    procedure ActionCopyUpdate(Sender: TObject);
    procedure ActionPasteUpdate(Sender: TObject);
    procedure ActionPasteExecute(Sender: TObject);
    procedure ActionCutUpdate(Sender: TObject);
    procedure ActionCutExecute(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ListRecordsData(Sender: TObject; Item: TListItem);
    procedure ListRecordsDataFind(Sender: TObject; Find: TItemFind;
      const FindString: String; const FindPosition: TPoint;
      FindData: Pointer; StartIndex: Integer; Direction: TSearchDirection;
      Wrap: Boolean; var Index: Integer);
    procedure edHexSelectionChange(Sender: TObject);
  protected
    DataRecords: TRecordList;
    SearchDialog: TSearchDialog;
    EditDialog: TEditDialog;

    IsBiff: boolean;

    procedure UpdateRecordCount;
    function SectionBegins(const aRecord: PRecord): boolean;virtual;
    function SectionEnds(const aRecord: PRecord): boolean;virtual;
    function GetIdName(const Id: word): string;virtual;abstract;

    procedure FixStreamOffsets(const First: integer);
  private
    procedure MoveSelected(const Dest: integer);
    procedure SetListColor(const s: string; const Ed: TEdit;
      const Color: TColor);
    function GetSelectedString: string;
    function LastSelected: integer;
    procedure FillHexText(const s: string);
    { Private declarations }
  public
    { Public declarations }
  end;

implementation
{$R *.DFM}

procedure TRecordEdit.UpdateRecordCount;
begin
  ListRecords.Items.Count:=DataRecords.Count;
  Statusbar.Panels[PanelTotalRecords].Text:=Format(TxtStatus, [DataRecords.Count]);
end;

function TRecordEdit.LastSelected:integer;
var
  i, Total:integer;
begin
  if ListRecords.Selected=nil then begin;result:=0;exit;end;
  Total:=1;Result:=ListRecords.Selected.Index;
  //Small optimization for continue selections...
  for i:= Result+1 to ListRecords.Items.Count do
  begin
   if Total>= ListRecords.SelCount then Break;
   if ListRecords.Items[i].Selected then begin inc(Total);Result:=i;end;
  end;
end;

procedure TRecordEdit.FillHexText(const s: string);
{const
  lh=8*3;
var
  i: integer;
}begin
  edHex.Text:=s;
{  for i:=0 to (length(s)-1) div lh do
  begin
    edHex.SelStart := i*lh;
    edHex.SelLength:= lh;
    if i mod 2= 0 then
      edHex.SelAttributes.color:=clblue else
      edHex.SelAttributes.Color:=clblack;
  end;
}end;

procedure TRecordEdit.ListRecordsSelectItem(Sender: TObject; Item: TListItem;
  Selected: Boolean);
begin
  if Selected then FillHexText(DataRecords[Item.Index].HexData) else edHex.Text:='';
  if ListRecords.SelCount>1 then StatusBar.Panels[PanelCurrentRecord].Text:= Format (TxtSelection,[ListRecords.SelCount])
  else if ListRecords.SelCount=1 then StatusBar.Panels[PanelCurrentRecord].Text:= Format (TxtCurrentRecord,[ListRecords.Selected.Index])
  else StatusBar.Panels[PanelCurrentRecord].Text:='';
end;

procedure TRecordEdit.FormCreate(Sender: TObject);
begin
  DataRecords:= TRecordList.Create;
  SearchDialog:= TSearchDialog.Create(Self);
end;

procedure TRecordEdit.SetListColor(const s:string; const Ed: TEdit; const Color: TColor);
begin
  if (UpperCase(TrimWoZero(s))= UpperCase(TrimWoZero(Ed.Text))) then ListRecords.Canvas.Brush.Color:=Color;
end;

procedure TRecordEdit.ListRecordsCustomDrawItem(Sender: TCustomListView;
  Item: TListItem; State: TCustomDrawState; var DefaultDraw: Boolean);
begin
    SetListColor(Item.Caption,SearchDialog.IdName,SearchDialog.Shape1.Brush.Color);
    SetListColor(Item.SubItems[0],SearchDialog.Id,SearchDialog.Shape1.Brush.Color);
    SetListColor(Item.Caption,SearchDialog.Edit1,SearchDialog.Shape2.Brush.Color);
    SetListColor(Item.SubItems[0],SearchDialog.Edit2,SearchDialog.Shape2.Brush.Color);
    SetListColor(Item.Caption,SearchDialog.Edit3,SearchDialog.Shape3.Brush.Color);
    SetListColor(Item.SubItems[0],SearchDialog.Edit4,SearchDialog.Shape3.Brush.Color);
    SetListColor(Item.Caption,SearchDialog.Edit5,SearchDialog.Shape4.Brush.Color);
    SetListColor(Item.SubItems[0],SearchDialog.Edit6,SearchDialog.Shape4.Brush.Color);
end;

procedure TRecordEdit.ListRecordsDragOver(Sender, Source: TObject; X, Y: Integer;
  State: TDragState; var Accept: Boolean);
begin
  Accept:=Source=ListRecords;
end;

procedure TRecordEdit.ListRecordsDragDrop(Sender, Source: TObject; X,
  Y: Integer);
var
  LI: TListItem;
begin
  LI:=ListRecords.GetItemAt(X,Y);
  if LI=nil then exit;
  MoveSelected(LI.Index);
end;

procedure TRecordEdit.MoveSelected(const Dest: integer);
var
  i,dx,sx:integer;
  TempList: TList;
  TempItem: PRecord;
  WaitCursor: IWaitCursor;
begin
  WaitCursor:=TWaitCursor.Create;
  if (Dest<0)or(ListRecords.Selected=nil)or (ListRecords.Items[Dest].Selected) then exit;
  try
    TempList:=TList.Create;
    try
      ListRecords.Items.BeginUpdate;
      try
        dx:=Dest;
        sx:=ListRecords.Selected.Index;
        for i:=LastSelected downto sx do
          if ListRecords.Items[i].Selected then
          begin
            new(TempItem);
            try
              CopyRecord(DataRecords[i], TempItem);
              TempList.Add(TempItem);
            except
              Dispose(TempItem);
              raise;
            end;
            ListRecords.Items[i].Selected:=false;
            DataRecords.Delete(i);
            if i<dx then dec(dx);
          end;

        for i:=0 to TempList.Count-1 do
          DataRecords.Insert(dx, TempList[i]);
        for i:=0 to TempList.Count-1 do ListRecords.Items[dx+i].Selected:=true;
        if dx<sx then FixStreamOffsets(dx-1) else FixStreamOffsets(sx-1);
      finally
        ListRecords.Items.EndUpdate;
      end; //finaly
    finally
      FreeAndNil(TempList);
    end; //finally
  finally
    UpdateRecordCount;
  end; //finally
end;

procedure TRecordEdit.ActionExportExecute(Sender: TObject);
var
  i:integer;
  f: TextFile;
  WaitCursor: IWaitCursor;
begin
  if not ExportDialog.Execute then exit;

  WaitCursor:=TWaitCursor.Create;
  AssignFile(f, ExportDialog.Filename);
  Rewrite(f);
  try
    for i:=0 to DataRecords.Count-1 do
      writeln(f,DataRecords[i].IdName,#9, Format('%4.4x',[DataRecords[i].id]),#9,
                DataRecords[i].Size,#9,{Format('%8x',[DataRecords[i].StreamPos],#9,}
                DataRecords[i].Data,#9,DataRecords[i].HexData);
  finally
    CloseFile(f);
  end; //finally
end;

procedure TRecordEdit.ActionDeleteExecute(Sender: TObject);
var
  i,First:integer;
  WaitCursor: IWaitCursor;
begin
  WaitCursor:=TWaitCursor.Create;
  if ListRecords.Selected=nil then exit;
  ListRecords.Items.BeginUpdate;
  try
    First:=ListRecords.Selected.Index;
    for i:= LastSelected downto First do
      if ListRecords.Items[i].Selected then
      begin
        DataRecords.Delete(i);
        if (i>First) and (i<DataRecords.Count) then ListRecords.Items[i].Selected:=false;
      end;
    FixStreamOffsets(First-1);
  finally
    UpdateRecordCount;
    ListRecords.Items.EndUpdate;
  end; //finaly
end;

procedure TRecordEdit.FixStreamOffsets(const First: integer);
var
  i:integer;
  StreamPos: int64;
  StrOfs:string;
  WaitCursor: IWaitCursor;
begin
  WaitCursor:=TWaitCursor.Create;
  if First<0 then begin
    StreamPos:=0;
    StrOfs:='';
  end else
  begin
    StreamPos:=DataRecords[First].Offset;
    StrOfs:='';i:=1;
    while(i<=Length(DataRecords[First].IdName))do
     if (DataRecords[First].IdName[i]=' ') then begin; StrOfs:=StrOfs+' ';inc(i); end else break;
  end;

  for i:=First+1 to DataRecords.Count-1 do
  begin
    if (i>0) then
    begin
      if SectionEnds( DataRecords[i-1]) and (length(StrOfs)-1>0) then Delete(StrOfs,Length(StrOfs)-1,2); //eof;
      if SectionBegins( DataRecords[i-1]) then StrOfs:=StrOfs+'  '; //bof;
      StreamPos:=StreamPos+SizeOf(TRecordHeader)+DataRecords[i-1].Size;
    end;
    DataRecords[i].Offset:= StreamPos;
    DataRecords[i].IdName:=StrOfs+Trim(DataRecords[i].IdName);
  end;

  ListRecords.Invalidate;
end;

procedure TRecordEdit.ActionEditExecute(Sender: TObject);
var
  Index: integer;
begin
  if ListRecords.Selected= nil then exit;

  if EditDialog=nil then EditDialog:= TEditDialog.Create(Self);
  EditDialog.edVersion.Visible:= not isBiff;
  EditDialog.lblVersion.Visible:= not isBiff;

  Index:=ListRecords.Selected.Index;
  EditDialog.edId.Text:=Format('%4.4x',[DataRecords[Index].Id]);
  EditDialog.edVersion.Text:=Format('%4.4x',[DataRecords[Index].Version]);
  EditDialog.edSize.Text:=IntToStr(DataRecords[Index].Size);
  EditDialog.edData.Text:=DataRecords[Index].HexData;
  if EditDialog.ShowModal<>mrOk then exit;
  DataRecords[Index].IdName:=GetIdName(StrToInt('$'+EditDialog.edId.Text));
  DataRecords[Index].Id:= StrToInt('$'+Trim(EditDialog.edId.Text));
  DataRecords[Index].Size:= StrToInt(EditDialog.edSize.Text);
  DataRecords[Index].Data:= SkipZero (HexToStr(EditDialog.edData.Text,StrToInt(EditDialog.edSize.Text)));
  DataRecords[Index].HexData:= Uppercase(EditDialog.edData.Text);
  DataRecords[Index].Version:=StrToInt('$'+Trim(EditDialog.edVersion.Text));
  ListRecordsSelectItem( Self, ListRecords.Selected, true);
  FixStreamOffsets(ListRecords.Selected.Index-1);
end;

procedure TRecordEdit.ActionEditUpdate(Sender: TObject);
begin
  ActionEdit.Enabled:=ListRecords.SelCount=1;
end;

procedure TRecordEdit.ActionMoveUpExecute(Sender: TObject);
var
  WaitCursor: IWaitCursor;
begin
  WaitCursor:=TWaitCursor.Create;

  if (ListRecords.Selected<>nil)and (ListRecords.Selected.Index>0) then
    MoveSelected(ListRecords.Items[ListRecords.Selected.Index-1].Index);
end;

procedure TRecordEdit.ActionMoveUpUpdate(Sender: TObject);
begin
  ActionMoveUp.Enabled:=(ListRecords.Selected<>nil) and (ListRecords.Selected.Index>0);
end;

procedure TRecordEdit.ActionDeleteUpdate(Sender: TObject);
begin
  ActionDelete.Enabled:=(ListRecords.Selected<>nil);
end;

procedure TRecordEdit.ActionMoveDownExecute(Sender: TObject);
var
  i:integer;
  WaitCursor: IWaitCursor;
begin
  WaitCursor:=TWaitCursor.Create;
  if (ListRecords.Selected<>nil)and not(ListRecords.Items[ListRecords.Items.Count-1].Selected)then
  begin
    i:=LastSelected;
    while (i>0) and not ListRecords.Items[i].Selected do dec(i);
    if i+2<=ListRecords.Items.Count then MoveSelected(ListRecords.Items[i+2].Index);
  end;
end;

procedure TRecordEdit.ActionMoveDownUpdate(Sender: TObject);
begin
  ActionMoveDown.Enabled:=(ListRecords.Selected<>nil) and not (ListRecords.Items[ListRecords.Items.Count-1].Selected);
end;

procedure TRecordEdit.ActionSearchExecute(Sender: TObject);
begin
  SearchDialog.ShowModal;
  ListRecords.Invalidate;
end;


procedure TRecordEdit.ActionExportUpdate(Sender: TObject);
begin
  ActionExport.Enabled:=DataRecords.Count>0;
end;

procedure TRecordEdit.ListRecordsDblClick(Sender: TObject);
begin
  ActionEdit.Execute;
end;

procedure TRecordEdit.ActionSearchUpdate(Sender: TObject);
begin
  ActionSearch.Enabled:=DataRecords.Count>0;
end;

function TRecordEdit.GetSelectedString: string;
var
  i: integer;
begin
  Result:='';
  if ListRecords.Selected=nil then exit;
  for i:=ListRecords.Selected.Index to LastSelected do
    if ListRecords.Items[i].Selected then
    begin
      Result:= Result+
               DataRecords[i].IdName+#9+ Format('%4.4x',[DataRecords[i].id])+#9;
      if not IsBiff then Result:=Result+ Format('%4.4x',[DataRecords[i].Version])+#9;
      Result:= Result+
               IntToStr(DataRecords[i].Size)+#9+Format('%8x',[DataRecords[i].Offset])+#9+
               DataRecords[i].Data+#9+DataRecords[i].HexData+#13#10;
    end;
end;

procedure TRecordEdit.ActionCopyExecute(Sender: TObject);
var
  WaitCursor: IWaitCursor;
begin
  WaitCursor:=TWaitCursor.Create;

  if ListRecords.Selected=nil then exit;
  Clipboard.AsText:=GetSelectedString;
end;

procedure TRecordEdit.ActionCopyUpdate(Sender: TObject);
begin
  ActionCopy.Enabled:= ListRecords.Selected<>nil;
end;

procedure TRecordEdit.ActionPasteUpdate(Sender: TObject);
begin
  ActionPaste.Enabled:= Clipboard.HasFormat(CF_TEXT) and (DataRecords.Count>0) and (ListRecords.Selected<>nil);
end;

procedure TRecordEdit.ActionPasteExecute(Sender: TObject);
var
  s: string;
  Posi, i: integer;
  MyId, MyVersion: word; MySize: int64;
  HexData: string;
  InsPos, FirstPos: integer;
  DataRec: PRecord;
  WaitCursor: IWaitCursor;
begin
  WaitCursor:=TWaitCursor.Create;

  if not Clipboard.HasFormat(CF_TEXT) then exit;
  if ListRecords.Selected=nil then exit;
  InsPos:= ListRecords.Selected.Index;
  FirstPos:=InsPos;
  s:=Clipboard.AsText;
  Posi:=1;
  try
    while ParseRecord(s,MyId, MyVersion, MySize, HexData, Posi, IsBiff) do
    begin
      New(DataRec);
      try
        DataRec.Id:=MyId;
        DataRec.IdName:= GetIdName(MyId);
        DataRec.Version:=MyVersion;
        DataRec.Size:=MySize;
        DataRec.Offset:=0;
        DataRec.Data:=SkipZero(HexToStr(HexData,MySize));
        DataRec.HexData:=HexData;
        DataRecords.Insert(InsPos,DataRec);
        inc(InsPos)
      except
        Dispose(DataRec);
        raise;
      end; //except
    end; //while
  finally
    UpdateRecordCount;
  end; //finally

  for i:=InsPos+1 downto FirstPos do  //This should be done in all the treeview, but is slow
    ListRecords.Items[i].Selected:=(i>=FirstPos) and (i<InsPos);
  FixStreamOffsets(FirstPos-1);
end;

procedure TRecordEdit.ActionCutUpdate(Sender: TObject);
begin
  ActionCut.Enabled:=ListRecords.Selected<>nil;

end;

procedure TRecordEdit.ActionCutExecute(Sender: TObject);
begin
  ActionCopy.Execute;
  ActionDelete.Execute;
end;

procedure TRecordEdit.FormDestroy(Sender: TObject);
begin
  FreeAndNil(DataRecords);
end;

procedure TRecordEdit.ListRecordsData(Sender: TObject; Item: TListItem);
begin
  if Screen.Cursor=crhourglass then exit;
  Item.Caption:= DataRecords[Item.Index].IdName;
  Item.SubItems.Add(Format('%4.4x',[DataRecords[Item.Index].Id]));
  if not IsBiff then Item.SubItems.Add(Format('%4.4x',[DataRecords[Item.Index].Version]));
  Item.SubItems.Add(IntToStr(DataRecords[Item.Index].Size));
  Item.SubItems.Add(Format('%8x',[DataRecords[Item.Index].Offset]));
  Item.SubItems.Add(DataRecords[Item.Index].Data);
  Item.SubItems.Add(DataRecords[Item.Index].HexData);
end;

procedure TRecordEdit.ListRecordsDataFind(Sender: TObject; Find: TItemFind;
  const FindString: String; const FindPosition: TPoint; FindData: Pointer;
  StartIndex: Integer; Direction: TSearchDirection; Wrap: Boolean;
  var Index: Integer);
var
  I: Integer;
  Found: Boolean;
begin
  I := StartIndex;
  if (Find = ifExactString) or (Find = ifPartialString) then
  begin
    repeat
      if (I >= DataRecords.Count) then
        if Wrap then I := 0 else Exit;
      Found := Pos(UpperCase(FindString), UpperCase(Trim(DataRecords[I].IdName))) = 1;
      Inc(I);
    until Found or (I = StartIndex);
    if Found then Index := I-1;
  end;
end;

function TRecordEdit.SectionBegins(const aRecord: PRecord): boolean;
begin
  Result:=false;
end;

function TRecordEdit.SectionEnds(const aRecord: PRecord): boolean;
begin
  Result:=false;
end;

procedure TRecordEdit.edHexSelectionChange(Sender: TObject);
begin
  StatusBar.Panels[PanelCursorPos].Text:= format (TxtCursorPos, [edHex.SelStart div 3]);
end;

end.
